@@ -25,6 +25,7 @@ import javax.websocket.ClientEndpoint;
 import javax.websocket.CloseReason;
 import javax.websocket.OnClose;
 import javax.websocket.Session;
+import javax.websocket.server.ServerEndpointConfig;
 
 import java.io.IOException;
 import java.net.URI;
@@ -39,7 +40,6 @@ import org.junit.Test;
 import org.junit.runner.RunWith;
 import org.xnio.FutureResult;
 
-
 import io.netty.buffer.Unpooled;
 import io.netty.handler.codec.http.websocketx.BinaryWebSocketFrame;
 import io.netty.handler.codec.http.websocketx.TextWebSocketFrame;
@@ -82,7 +82,7 @@ public class AnnotatedEndpointTest {
                 .addServletContextAttribute(WebSocketDeploymentInfo.ATTRIBUTE_NAME,
                         new WebSocketDeploymentInfo()
                                 .setBuffers(DefaultServer.getBufferPool())
-                                .setWorker(DefaultServer.getWorker())
+                                .setWorker(DefaultServer.getWorkerSupplier())
                                 .addEndpoint(MessageEndpoint.class)
                                 .addEndpoint(AnnotatedClientEndpoint.class)
                                 .addEndpoint(AnnotatedClientEndpointWithConfigurator.class)
@@ -100,6 +100,10 @@ public class AnnotatedEndpointTest {
                                         deployment = container;
                                     }
                                 })
+                                .addEndpoint(ServerEndpointConfig.Builder.create(
+                                        AnnotatedAddedProgrammaticallyEndpoint.class,
+                                        AnnotatedAddedProgrammaticallyEndpoint.PATH)
+                                        .build())
                 )
                 .addServlet(new ServletInfo("redirect", RedirectServlet.class)
                 .addMapping("/redirect"))
@@ -130,6 +134,18 @@ public class AnnotatedEndpointTest {
         client.destroy();
     }
 
+    @Test
+    public void testStringOnMessageAddedProgramatically() throws Exception {
+        final byte[] payload = "foo".getBytes();
+        final FutureResult latch = new FutureResult();
+
+        WebSocketTestClient client = new WebSocketTestClient(WebSocketVersion.V13, new URI("ws://" + DefaultServer.getHostAddress("default") + ":" + DefaultServer.getHostPort("default") + "/ws/programmatic"));
+        client.connect();
+        client.send(new TextWebSocketFrame(Unpooled.wrappedBuffer(payload)), new FrameChecker(TextWebSocketFrame.class, "oof".getBytes(), latch));
+        latch.getIoFuture().get();
+        client.destroy();
+    }
+
     @Test
     public void testRedirectHandling() throws Exception {
         AnnotatedClientEndpoint.reset();
@@ -181,6 +197,7 @@ public class AnnotatedEndpointTest {
 
     @Test
     public void testCloseReason() throws Exception {
+        AnnotatedClientEndpoint.reset();
         MessageEndpoint.reset();
 
         Session session = deployment.connectToServer(AnnotatedClientEndpoint.class, new URI("ws://" + DefaultServer.getHostAddress("default") + ":" + DefaultServer.getHostPort("default") + "/ws/chat/Bob"));
@@ -352,6 +369,7 @@ public class AnnotatedEndpointTest {
     }
 
 
+
     @ClientEndpoint
     public static class DoNothingEndpoint {}
 
